/*
 * Copyright: GNU Public License 2 applies
 *
 * CDDA2WAV
 *
 * last change:
 *   18.12.93 - first version,	OK
 *   01.01.94 - generalized & clean up HE
 *   10.06.94 - first linux version HE
 *   12.06.94 - wav header alignment problem fixed HE
 *
 * TODO: 
 *       support pipes
 *       test program with multimedia cds (have none)
 *       
 */

#define VERSION	"0.0"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "uti.h"
#ifdef ASPI
#include "aspidef.h"
#else
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#endif

#define DEBUG(x)	
#undef DEBUGMODE

#define MAXTRK	30	/* maximum of audio tracks */

#define CB_CDDASECTOR	2368 /* including sub q channel */
#define CB_QSUBCHANNEL	16
#define CB_CDROMSECTOR	2048 /* sector size of data cds */
#define CB_AUDIO	(CB_CDDASECTOR-CB_QSUBCHANNEL) /* excluding sub q channel */

typedef struct TOC {	/* structure of table of contents (cdrom) */
  BYTE _reserved1;
  BYTE bFlags;
  BYTE bTrack;
  BYTE _reserved2;
  DWORD dwStartSector;
} TOC;

static BYTE dev_name [200] = CD_DEVICE;		/* device name */

static BYTE bufferCdRom [CB_CDROMSECTOR];
static BYTE bufferCdda [NSECTORS * CB_CDDASECTOR];
static BYTE bufferAudio [NSECTORS * CB_AUDIO];

static BYTE g_track=0xff, g_index=0xff;		/* current track, index */
static BYTE g_minute=0xff, g_seconds=0xff;	/* curr. minute, second */

static int nTocEntries;
static TOC g_toc [MAXTRK]; /* 100 */

static void EnableAudioMode (BOOL fAudioMode);

/*
 * ---------------------------------------------------------------------
 *  definitions for RIFF-output (from windows MMSYSTEM)
 * ---------------------------------------------------------------------
 */

typedef DWORD FOURCC;	/* a four character code */

typedef struct CHUNKHDR {
  FOURCC ckid;		/* chunk ID */
  DWORD dwSize; 	/* chunk size */
} CHUNKHDR;

/* flags for 'wFormatTag' field of WAVEFORMAT */
#define WAVE_FORMAT_PCM 1

/* specific waveform format structure for PCM data */
typedef struct pcmwaveformat_tag {
  WORD wFormatTag;	/* format type */
  WORD nChannels;	/* number of channels (i.e. mono, stereo, etc.) */
  DWORD nSamplesPerSec; /* sample rate */
  DWORD nAvgBytesPerSec;/* for buffer estimation */
  WORD nBlockAlign;	/* block size of data */
  WORD wBitsPerSample;
} PCMWAVEFORMAT;
typedef PCMWAVEFORMAT *PPCMWAVEFORMAT;


/* MMIO macros */
#define mmioFOURCC(ch0, ch1, ch2, ch3) \
  ((DWORD)(BYTE)(ch0) | ((DWORD)(BYTE)(ch1) << 8) | \
  ((DWORD)(BYTE)(ch2) << 16) | ((DWORD)(BYTE)(ch3) << 24))

#define FOURCC_RIFF	mmioFOURCC ('R', 'I', 'F', 'F')
#define FOURCC_LIST	mmioFOURCC ('L', 'I', 'S', 'T')
#define FOURCC_WAVE	mmioFOURCC ('W', 'A', 'V', 'E')
#define FOURCC_FMT	mmioFOURCC ('f', 'm', 't', ' ')
#define FOURCC_DATA	mmioFOURCC ('d', 'a', 't', 'a')


/* simplified Header for standard WAV files */
typedef struct WAVEHDR {
  CHUNKHDR chkRiff;
  FOURCC fccWave;
  CHUNKHDR chkFmt;
  WORD wFormatTag;	/* format type */
  WORD nChannels;	/* number of channels (i.e. mono, stereo, etc.) */
  DWORD nSamplesPerSec; /* sample rate */
  DWORD nAvgBytesPerSec;/* for buffer estimation */
  WORD nBlockAlign;	/* block size of data */
  WORD wBitsPerSample;
  CHUNKHDR chkData;
} WAVEHDR;

#define IS_STD_WAV_HEADER(waveHdr) ( \
  waveHdr.chkRiff.ckid == FOURCC_RIFF && \
  waveHdr.fccWave == FOURCC_WAVE && \
  waveHdr.chkFmt.ckid == FOURCC_FMT && \
  waveHdr.chkData.ckid == FOURCC_DATA && \
  waveHdr.wf.wFormatTag == WAVE_FORMAT_PCM)

static FILE *audio;


#ifdef	ASPI
/*
 * ASPI (Advanced SCSI Programming Interface) stuff
 */

static SRB_IO srb;


static BOOL AspiPrintError (int err)
{
  if (err == ASPIE_SUCCESS) return FALSE;
  printf ("cdrom: %s\n", AspiGetErrorText (err));
  return TRUE;
}
#else

/*  ioctl scsi interface. os specific.
 *
 *  Currently all cdrom scsi commands are given explicit via the 
 *  SCSI_IOCTL_SEND_COMMAND ioctl.
 *
 *  When a future kernel has these commands incorporated, an other
 *  interface may be used.
 */

static int cd_fd;


static int handle_scsi_cmd( unsigned size1, unsigned char *cmd, 
		      unsigned size2, unsigned char *data, 
		      unsigned size3, unsigned char *buffer )
{
    int status;

    /* TODO: get rid off this extra copying to/from ownbuff */
    static unsigned char ownbuff[20+NSECTORS*CB_CDDASECTOR];

    *( (int *) ownbuff) = size2;	/* input buffer length */
    *( ((int *) ownbuff) +1 ) = size3;	/* output buffer length */

    if (size1 && cmd) memcpy(ownbuff+2*sizeof(int), cmd, size1);
    if (size2 && data) memcpy(ownbuff+2*sizeof(int)+size1, data, size2);

#if	0
    { int i; for (i=0; i< 2*sizeof(int)+size1 + size2; i++) {
	printf("0x%02x    ",ownbuff[i]);
    }} printf("\n");
#endif

    status = ioctl( cd_fd, 1 /* SCSI_IOCTL_SEND_COMMAND */, ownbuff );

    /* get output back */
    if (size3 && buffer) memcpy(buffer, ownbuff+2*sizeof(int), size3);
    if (status) {
	printf( "\nioctl(SCSI_IOCTL_SEND_COMMAND) status\t= 0x%x, cmd=%x\n", 
	       status, cmd[0] );
    }
    return status;
}
#endif

/* actually reverse. For conversion on low endian machines */
static VOID *Swap (VOID *p, int size)
{
  char *pc = p;
  char tmp;

  if (size == 4) {
    tmp = pc [0];	/* reverse 4 bytes */
    pc [0] = pc [3];
    pc [3] = tmp;
    tmp = pc [1];
    pc [1] = pc [2];
    pc [2] = tmp;
  } else {
    tmp = pc [0];	/* reverse 2 bytes */
    pc [0] = pc [1];
    pc [1] = tmp;
  }
  return p;
}

static WAVEHDR waveHdr;
static DWORD nBytesDone = 0;

/* define size-related entries in wave header, update and close file */
static VOID CloseAudio (void)
{
  DWORD cbData = nBytesDone;/* / 4
                 / (44100 / waveHdr.wf.nSamplesPerSec)
                 * waveHdr.wf.nChannels 
                 * ((waveHdr.wBitsPerSample + 7) / 8);*/

  fseek(audio, 0L, SEEK_SET);

  waveHdr.chkRiff.dwSize = cbData + sizeof(WAVEHDR) - sizeof(CHUNKHDR) ;
  waveHdr.chkData.dwSize = cbData;

  fwrite (&waveHdr, sizeof (waveHdr), 1, audio);

  fclose (audio);
}

/* Close audio file with correct header (length) and 
   set cdrom drive to normal mode back. */
static VOID Cleanup (void)
{
  CloseAudio ();
  EnableAudioMode (FALSE);
}


static VOID FatalError (char *szMessage, ...)
{
  vprintf (szMessage, (char *)(&szMessage + 1));
  Cleanup();
  exit (1);
}


/* open the wav output file and prepare the wav header. */
static VOID OpenAudio (char *fname, unsigned rate, int nBitsPerSample, 
		       int channels)
{
  audio = fopen (fname, "wb");

  waveHdr.chkRiff.ckid = FOURCC_RIFF;
  waveHdr.fccWave = FOURCC_WAVE;
  waveHdr.chkFmt.ckid = FOURCC_FMT;
  waveHdr.chkFmt.dwSize = sizeof (PCMWAVEFORMAT);
  waveHdr.wFormatTag = WAVE_FORMAT_PCM;
  waveHdr.nChannels = channels;
  waveHdr.nSamplesPerSec = rate;
  waveHdr.nBlockAlign = channels * ((nBitsPerSample + 7) / 8);
  waveHdr.nAvgBytesPerSec = waveHdr.nBlockAlign * waveHdr.nSamplesPerSec;
  waveHdr.wBitsPerSample = nBitsPerSample;
  waveHdr.chkData.ckid = FOURCC_DATA;

  fwrite (&waveHdr, sizeof (waveHdr), 1, audio);
}


/* open the cdrom device */
void OpenCdRom (void)
{
#ifndef	ASPI

  cd_fd = open(dev_name,O_RDWR);
  if (cd_fd < 0)
    FatalError ("No access to %s.\n",dev_name);
  /* no checks so far. TODO at least a cdrom drive should be ensured. */

#else  
  /* ASPI variant */
  while (targetId < 7 && AspiGetDeviceType (0, targetId, 0) != DTC_CDROM)
    targetId++;
  if (targetId == 7)
    FatalError ("No CD-ROM drive found.\n");

  memzero (srb);
  srb.srb.bCmd = SRBC_EXECIO;
  srb.srb.bAdapterId = 0;
  srb.srb.bFlags = 0*SRBF_READ;
  srb.bTargetId = (BYTE)targetId;
  srb.bLun = 0;
  srb.lpData = NULL;
  srb.cbData = 0;
  srb.cbSense = 14;
  srb.cCdb = 6;
#endif
}

/* Read NSECTORS of cdda sectors to bufferCdda */
static void ReadCdRom (LONG lSector)
{
  /* READ10, flags, block1 msb, block2, block3, block4 lsb, reserved, 
     transfer len msb, transfer len lsb, block addressing mode */
  static BYTE cmd [10] = {0x28, 0, 0, 0, 0, 0, 0, 0, 0, 0};

  cmd [2] = (BYTE)(lSector >> 24);
  cmd [3] = (BYTE)((lSector >> 16) & 0xFF);
  cmd [4] = (BYTE)((lSector >> 8) & 0xFF);
  cmd [5] = (BYTE)(lSector & 0xFF);
  cmd [8] = (BYTE)NSECTORS;

#ifdef ASPI
  srb.lpData = bufferCdda;
  srb.cbData = NSECTORS * CB_CDDASECTOR;
  srb.cCdb = 10;
  memcpy (&srb.cdb, cmd, 10);

{ int err = AspiRequestWait (&srb.srb);
  if (AspiPrintError (err))
    FatalError ("Read CD-ROM failed");
}
#else
  if (handle_scsi_cmd(sizeof(cmd), cmd, 0, NULL, 
		      NSECTORS * CB_CDDASECTOR, bufferCdda ))
    FatalError ("Read CD-ROM failed");
#endif
}

static int quiet = 0;
static int verbose = 0;

/* print the track currently read */
static void UpdateTrackData (short p_num)
{
  if (!quiet) { printf ("\ntrack: %.2X, ", p_num); fflush(stdout); }
  g_track = (char)p_num;
}


/* print the index currently read */
static void UpdateIndexData (short p_num)
{
  if (!quiet) { printf ("index: %.2X\n", p_num); fflush(stdout); }
  g_index = (char)p_num;
}


/* print the time of track currently read */
static void UpdateTimeData (short p_min, short p_sec)
{
  if (!quiet) { printf ("time: %.2X:%.2X\r", p_min, p_sec); fflush(stdout); }
  g_minute = (char)p_min;
  g_seconds = (char)p_sec;
}

static VOID AnalyzeQchannel ( char *pSrc )
{
      /* analyze Q-sub channel data */
      if (pSrc [1] == 1) {
        if (((BYTE *)pSrc) [2] != g_track)
	  UpdateTrackData (((BYTE *)pSrc) [2]);
        if (((BYTE *)pSrc) [3] != g_index)
	  UpdateIndexData (((BYTE *)pSrc) [3]);
	if (((BYTE *)pSrc) [5] != g_seconds ||
      	    ((BYTE *)pSrc) [4] != g_minute)
	  UpdateTimeData (((BYTE *)pSrc) [4], ((BYTE *)pSrc) [5]);
      }
}

static int waitforsignal = 0;

/* convert cdda data to required output format */
static int SaveBuffer (unsigned rate, int nBitsPerSample, int channels, 
                 LONG SecsToDo, DWORD *BytesDone)
{
  short i;
  char *pSrc = (char *)bufferCdda;
  BYTE *pDst = (BYTE *)bufferAudio;
  short nSamplesLeftInSector = ((CB_CDDASECTOR-CB_QSUBCHANNEL)/4);
  unsigned undersampling = 44100 / rate;
  int sh_bits = 16 - nBitsPerSample;
  unsigned BytesToDo = (SecsToDo > NSECTORS ? NSECTORS : SecsToDo) * CB_CDDASECTOR;
  BYTE *pStart = pDst;
  static int any_signal = 0;
  char *pSrcStop = pSrc + BytesToDo;

  while (pSrc < pSrcStop) {
    long lsum = 0;	/* accumulate amplitudes in long ints */
    long rsum = 0;

    for (i=0; i < undersampling; i++) {
	    /* LSB l, MSB l, LSB r, MSB r, ... */
      lsum += *((signed short *)pSrc)++;	/* left channel */
      rsum += *((signed short *)pSrc)++;	/* right channel */
    }
    lsum /= undersampling;
    rsum /= undersampling;

    if (channels == 1) {
      short sum;       /* mono section */
      sum = (lsum + rsum) >> 1;
      if (nBitsPerSample == 8) {
        if ((BYTE)(sum >> 8) != 0) {
          if (!any_signal) pStart = pDst;
          any_signal = 1;
        }
        *pDst++ = (BYTE)(sum >> 8) + (1 << 7);	/* ok */
      } else {
        sum >>= sh_bits;
        if (sum != 0) {
          if (!any_signal) pStart = pDst;
          any_signal = 1;
        }
        *((short *)pDst)++ = sum; /* ok */
      }
    } else {
      /* stereo section */
      if (nBitsPerSample == 8) {
        if ((short)(lsum >> 8) != 0 || 
	    (short)(rsum >> 8) != 0) {
          if (!any_signal) pStart = pDst;
          any_signal = 1;
        }
        *pDst++ = (BYTE)(((short) lsum) >> 8) + (1 << 7);
        *pDst++ = (BYTE)(((short) rsum) >> 8) + (1 << 7);	/* ok */
      } else {
        if ((short)(lsum >> sh_bits) != 0 || 
	    (short)(rsum >> sh_bits) != 0) {
          if (!any_signal) pStart = pDst;
          any_signal = 1;
        }
        *((short *)pDst)++ = (((short) lsum) >> sh_bits);
        *((short *)pDst)++ = (((short) rsum) >> sh_bits); /* ok */
      }
    }

    nSamplesLeftInSector -= undersampling;
    if (nSamplesLeftInSector == 0) {
      if (!quiet)
	  AnalyzeQchannel(pSrc);

      /* skip Q-sub channel data */
      pSrc += CB_QSUBCHANNEL;
      nSamplesLeftInSector = ((CB_CDDASECTOR - CB_QSUBCHANNEL)/4);
    }
  } /* while */
  if (!waitforsignal) pStart = (BYTE *)bufferAudio;

  if (!waitforsignal || any_signal) {
    int retval = 0;
    if (!(retval = fwrite (pStart, pDst - pStart, 1, audio) != 1))
	*BytesDone += pDst - pStart;
    return retval;
  } else return 0;
}


static void EnableAudioMode (BOOL fAudioMode)
{

#ifdef	MODE_VIA_KERNEL
    if (ioctl( cd_fd, CDROMSETRAWAUDIO ))
       fprintf( stderr, "scsi_ioctl cdda/audio failed\n" );
#else
  /* reserved, Medium type=0, Dev spec Parm = 0, block descriptor len 0 oder 8,
     Density (cd format) 
     (0=YellowBook, XA Mode 2=81h, XA Mode1=83h and raw audio via SCSI=82h),
     # blks msb, #blks, #blks lsb, reserved,
     blocksize, blocklen msb, blocklen lsb,
   */
  static BYTE mode [12] = {0, 
                           0, 0, 
                           8,       /* Block Descriptor Length */
                           0,       /* Density Code */
                           0, 0, 0, /* # of Blocks */
                           0,       /* reserved */
                           0, 0, 0};/* Blocklen */

  /* MODE_SELECT, page = SCSI-2  save page disabled, reserved, reserved, 
     parm list len, flags */
  static BYTE cmd [6] = {0x15, /* MODE_SELECT */
			 0x10, /* no save page */
			    0, /* reserved */
			    0, /* reserved */
			    sizeof (mode), /* ?? */
			    0}; /* reserved */

  if (fAudioMode) {
    mode [4] = 0x82;  			/* cdda density */
    mode [10] = (CB_CDDASECTOR >> 8);   /* block length "msb" */
    mode [11] = (CB_CDDASECTOR & 0xFF);
  } else {
    mode [4] = 0x00; 			/* normal density */
    mode [10] = (CB_CDROMSECTOR >> 8);  /* block length "msb" */
    mode [11] = (CB_CDROMSECTOR & 0xFF); /* block length lsb*/
  }

#ifndef ASPI
  if (handle_scsi_cmd (sizeof(cmd), cmd, 
		       sizeof(mode), mode, 
		       0, NULL))
        perror ("Audio mode switch failed\n");
#else

  srb.lpData = mode;
  srb.cbData = sizeof (mode);
  srb.cCdb = 6;
  memcpy (&srb.cdb, cmd, 6);
  { 
    int err = AspiRequestWait (&srb.srb);
    if (AspiPrintError (err)) {
      if (err == ASPIE_UNITATTENTION)
        err = AspiRequestWait (&srb.srb);
      if (AspiPrintError (err))
        FatalError ("Enable audio mode failed\n");
    }
  }
#endif
#endif
}


static VOID ReadToc (void)
{
  /* READTOC, MSF format flag, res, res, res, res, Start track, len msb,
     len lsb, flags */
  static BYTE cmd [10] = {
    0x43, 0, 0, 0, 0, 0, 1, CB_CDROMSECTOR >> 8, CB_CDROMSECTOR & 0xFF, 0
  };
#ifdef ASPI
  srb.lpData = bufferCdRom;
  srb.cbData = CB_CDROMSECTOR;
  srb.cCdb = 10;
  memcpy (&srb.cdb, cmd, 10);

  {
    int err = AspiRequestWait (&srb.srb);
    if (AspiPrintError (err))
      FatalError ("Read TOC failed.");
  }
#else
  if (handle_scsi_cmd ( sizeof(cmd), cmd, 0, NULL, 
		       sizeof(bufferCdRom), bufferCdRom))
      FatalError ("Read TOC failed.");
#endif
  nTocEntries = ((bufferCdRom [0] << 8) + bufferCdRom [1] - 2) / 8;
  memcpy (g_toc, bufferCdRom + 4, 8 * nTocEntries);
}


static VOID DisplayToc (void)
{
  int i;
  DWORD dw;
  unsigned mins;
  float secnds;

  dw = g_toc [nTocEntries-1].dwStartSector;
  Swap (&dw, 4);
  mins= dw / (75L*60);
  secnds= fmod(dw , 60*75) / 75.0;
  /* g_toc [i].bFlags contains two fields:
       bits 7-4 (ADR) : 0 no sub-q-channel information
                      : 1 sub-q-channel contains current position
		      : 2 sub-q-channel contains media catalog number
		      : 3 sub-q-channel contains International Standard
		                                 Recording Code ISRC
		      : other values reserved
       bits 3-0 (Control) :
       bit 3 : when set indicates there are 4 audio channels else 2 channels
       bit 2 : when set indicates this is a data track else an audio track
       bit 1 : when set indicates digital copy is permitted else prohibited
       bit 0 : when set indicates pre-emphasis is present else not present
  */
  if (verbose) {
    printf("track pre-emphasis copy-permitted tracktype channels sub-Q-channel\n");
    for (i=0; i<nTocEntries; i++) {
      printf("%5u %12s %14s %18s ",i,
			g_toc [i].bFlags & 1 ? "  " : "no",
			g_toc [i].bFlags & 2 ? "  " : "no",
			g_toc [i].bFlags & 4 ? "data          " : 
			g_toc [i].bFlags & 8 ? "audio        4" :
	                                       "audio        2"
                       );
      switch (g_toc [i].bFlags >> 4) {
        case 0:  printf("no data\n"); break;
	case 1:  printf("position\n"); break;
	case 2:  printf("media catalog\n"); break;
	case 3:  printf("ISRC\n"); break;
	default: printf("invalid\n");
      }
    }
  }

  printf ("Table of Contents: total tracks:%d, (total time %u:%05.2f)\n",
          nTocEntries-1, mins, secnds);
  for (i=0; i<nTocEntries; i++) {
    if (g_toc [i].bFlags & 4) continue;	/* skip nonaudio tracks */
    if (g_toc [i].bTrack <= MAXTRK) {
      DWORD dw2 = g_toc [i].dwStartSector;
      dw = g_toc [i+1].dwStartSector;
      Swap (&dw, 4);
      Swap (&dw2, 4);
      mins= (dw - dw2) / (75L*60);
      secnds= fmod(dw-dw2 , 60*75) / 75.0;
      printf (" %d.(%u:%05.2f)", g_toc [i].bTrack,mins,secnds);
    }
    if (i < nTocEntries-1)
       if ((i+1) % 5 == 0) printf("\n");
       else putchar (',');
  }
  printf ("\n");
}


static LONG GetStartSector (int p_track)
{
  int i;

  for (i=0; i<nTocEntries; i++) {
    if (g_toc [i].bTrack == p_track) {
      DWORD dw = g_toc [i].dwStartSector;
      if (g_toc [i].bFlags & 4)
	return -1;
      Swap (&dw, 4);
      return dw;
    }
  }

  return -1;
}


static LONG GetEndSector (int p_track)
{
  int i;

  for (i=1; i<nTocEntries; i++) {
    if (g_toc [i-1].bTrack == p_track) {
      DWORD dw = g_toc [i].dwStartSector;
      if (g_toc [i].bFlags & 4)
	return -1;
      Swap (&dw, 4);
      return dw-1;
    }
  }

  return -1;
}



static int FirstTrack (void)
{
  int i;
  
  for (i=0; i<nTocEntries; i++) {
    if (g_toc [i].bTrack != 0xAA &&
	!(g_toc [i].bFlags & 4))
      return g_toc [i].bTrack;
  }
  return 0;
}

#if	0
static LONG NextTrack (int p_offset)
{
  int track = (g_track >> 4) * 10 + (g_track & 15);
  long res;

  track += p_offset;
  res = GetStartSector (track);
  return res == -1 ? GetStartSector (track - p_offset) : res;
}
#endif

static VOID usage(void)
{
	fprintf(stderr, "cdda2wav [-c chans] [-s] [-m] [-b bits] [-r rate] [-t track] [-o offset]\n");
	fprintf(stderr, "         [-D device] [-v] [wavfile.wav]; Version %s\n",VERSION);
	fprintf(stderr, "cdda2wav copies parts from audio cd's directly to WAV files.\n");
	fprintf(stderr, "It requires a toshiba XM3401 scsi cdrom drive"
#ifdef ASPI
                        " and an aspi driver"
#endif
		        " to work.\n");
	fprintf(stderr, "options: -c channels : set 1 for mono, or 2 for stereo recording.\n");
	fprintf(stderr, "         -D device   : set the cdrom device.\n");
	fprintf(stderr, "         -s          : set to stereo recording.\n");
	fprintf(stderr, "         -m          : set to mono recording.\n");
	fprintf(stderr, "         -b bits     : set bits per sample per channel (8, 12 or 16 bits).\n");
	fprintf(stderr, "         -r rate     : set rate in samples per second. Possible values are:\n");
	fprintf(stderr, "            44100,22050,14700,11025,7350,3675,3150,1575,900 Hertz.\n");
	fprintf(stderr, "         -t track    : select start track.\n");
	fprintf(stderr, "         -o offset   : start 'offset' sectors behind start track.\n");
	fprintf(stderr, "                       one sector equivalents 1/75 second.\n");
	fprintf(stderr, "         -d duration : set recording time in seconds or 0 for whole track.\n");
	fprintf(stderr, "         -w          : wait for signal, then start recording.\n");
	fprintf(stderr, "         -v          : print all information about the cd.\n");
	fprintf(stderr, "         -q          : quiet operation, no screen output.\n");
	fprintf(stderr, "defaults: mono, 16 bit, 22050 Hz, track 1, no offset, 16 seconds, 'audio.wav'\n"
                  "          don't wait for signal, not quiet, not verbose, use /dev/sr0\n");
  exit(1);
}

/* and finally: the MAIN program */
int main (int argc, char *argv [])
{
  int args = 0;
  LONG lSector;
  LONG lSector_p1;
  LONG sector_offset = 0;
  LONG nSectorsToDo = ((DURATION*75) / NSECTORS) * NSECTORS;
  LONG time = nSectorsToDo / 75;
  int track = 1;
  int	channels = CHANNELS;
  unsigned rate = 44100 / UNDERSAMPLING;
  int bits = BITS_P_S;
  char fname[100] = FILENAME;
#if	defined	(MSDOS) || defined(__MSDOS__)
#include <conio.h>
  int key_pressed = 0;
  LONG req_space;
#endif

  /* command options parsing */
  /* when created under dos, there was no getopt() */
  while (--argc) {
    args++;
    /* override device */
    if (!strncmp("-D",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	strcpy(dev_name,argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  strcpy(dev_name,argv[args]);
	  continue;
        } else usage();
      }
    }
    /* channels */
    if (!strncmp("-c",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	channels = atoi(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  channels = atoi(argv[args]);
	  continue;
        } else usage();
      }
    }
    /* bits */
    if (!strncmp("-b",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	bits = atoi(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  bits = atoi(argv[args]);
	  continue;
        } else usage();
      }
    }
    /* rate */
    if (!strncmp("-r",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	rate = atoi(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  rate = atoi(argv[args]);
	  continue;
        } else usage();
      }
    }
    /* start track */
    if (!strncmp("-t",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	track = atoi(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  track = atoi(argv[args]);
	  continue;
	} else usage();
      }
    }
    /* recording time */
    if (!strncmp("-d",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	time = atoi(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  time = atoi(argv[args]);
	  continue;
	} else usage();
      }
    }
    /* sector offset */
    if (!strncmp("-o",argv[args],2)) {
      if (strlen (argv[args]) > 2) {
	sector_offset = atol(argv[args]+2);
	continue;
      } else {
        args++;
        if (--argc) {
	  sector_offset = atol(argv[args]);
	  continue;
	} else usage();
      }
    }
    /* stereo */
    if (!strncmp("-s",argv[args],2)) {
      channels = 2;
      continue;
    }
    /* mono */
    if (!strncmp("-m",argv[args],2)) {
      channels = 1;
      continue;
    }
    /* wait for signal */
    if (!strncmp("-w",argv[args],2)) {
      waitforsignal = 1;
      continue;
    }
    /* verbose. can be overridden by quiet */
    if (!strncmp("-v",argv[args],2)) {
      verbose = 1;
      continue;
    }
    /* quiet */
    if (!strncmp("-q",argv[args],2)) {
      quiet = 1;
      verbose = 0;
      continue;
    }
    /* other options unknown */
    if (*argv[args] == '-') usage();

    /* filename given */
    strcpy(fname,argv[args]);

    /* should be the last parameter */
    if (argc > 1) usage();
  }

  /* check all parameters */
  if (channels != 1 && channels != 2) usage();
  if (bits != 8 && bits != 12 && bits != 16) usage();
  if (44100 % rate != 0) usage();
  if (track < 1 || track > 50) usage();
  if (time < 0 || time > 4440) usage();

#ifdef ASPI
  if (not AspiInit ()) {
    printf ("Unable to open ASPI device.\n");
    exit(1);
  }
#endif


  OpenCdRom ();
  EnableAudioMode (TRUE);

  signal( SIGINT, exit );

  atexit (Cleanup);

  ReadToc ();
  if (!quiet) DisplayToc ();

  if (!FirstTrack ())
    FatalError ("This is no audio disk\n");

  lSector = GetStartSector (track);
  lSector_p1 = GetEndSector (track) + 1;
  if (lSector < 0)
    FatalError ("track %d not found\n", track);

  lSector += sector_offset;
  /* check against end sector of track */
  if (lSector >= lSector_p1) {
    printf("sector offset exceeds track size (ignored)\n");
    lSector -= sector_offset;
  }

  if (time == 0) {
    /* set time to track time */
    time = ((lSector_p1 - lSector)) / 75;
  }

  remove(fname);


#if	defined	(MSDOS) || defined(__MSDOS__)
  /* calculate required disk space */
  req_space = sizeof(WAVEHDR) + (bits + 7)/8 * channels * (LONG)rate * time;

  /* check disk space */
  {
#include <dos.h>
#include <ctype.h>
    struct diskfree_t d;
    unsigned drive;
    LONG avail;

    if (fname[1] != ':') {
      /* get default */
      drive = 0;
    } else {
      /* get specified drive */
      drive = toupper(fname[0]) - '@';
    }
    _dos_getdiskfree(drive, &d);
    avail = (long) d.avail_clusters
           * (long) d.bytes_per_sector
           * (long) d.sectors_per_cluster;
    if (req_space > avail) {
      if (!quiet) fprintf(stderr, "unsufficient disk space! reducing recording length.\n");
      time = (avail - sizeof(WAVEHDR)) / ((bits + 7)/8 * channels*(LONG)rate);
      req_space = sizeof(WAVEHDR) + (bits + 7)/8 * channels*(LONG)rate * time;
      if (!quiet) fprintf(stderr, 
	"maximum possible are %ld seconds <=> %ld Bytes.\n",time,req_space);
    }
  }
#endif

  /* calculate # of sectors to read */
  nSectorsToDo = time * 75;

  if (!quiet) printf(
     "recording %lu seconds %s with %d bits @ %u Hz ->'%s'...\n"
     ,time ,channels == 1 ? "mono":"stereo", bits, rate, fname);

  OpenAudio (fname, rate, bits, channels);

  while (nSectorsToDo > 0 
#if	defined	(MSDOS) || defined(__MSDOS__)
	                   && !(key_pressed = kbhit())
#endif
                         ) {
    ReadCdRom (lSector);
    lSector += NSECTORS;
    if (SaveBuffer (rate, bits, channels, nSectorsToDo, &nBytesDone)) {
	fprintf (stderr, "\ndisk space exhausted. Stopped.\n");
	break;
    }
    nSectorsToDo -= NSECTORS;
  }

#if	defined	(MSDOS) || defined(__MSDOS__)
  if (key_pressed) {
    getch();
    if (!quiet) printf("\naborted.");
  }
#endif
  printf("\n");
  CloseAudio ();
  EnableAudioMode (FALSE);

  return 0;
}
